import numpy as np


class Layer:
    """Base class for layers"""

    def forward(self, x):
        raise NotImplementedError

    def backward(self, grad):
        raise NotImplementedError

    def parameters(self):
        return []



# 1. Linear Layer

class Linear(Layer):

    def __init__(self, in_features, out_features):

        self.W = np.random.randn(in_features, out_features) * np.sqrt(2 / in_features)
        self.b = np.zeros((1, out_features))

        # Initialize gradients
        self.dW = np.zeros_like(self.W)
        self.db = np.zeros_like(self.b)

    def forward(self, x):

        self.x = x
        return x @ self.W + self.b

    def backward(self, grad):

        self.dW = self.x.T @ grad
        self.db = np.sum(grad, axis=0, keepdims=True)

        dx = grad @ self.W.T

        return dx


# 2. Dropout

class Dropout(Layer):

    def __init__(self, p=0.5):
        self.p = p
        self.training = True

    def forward(self, x):

        if not self.training:
            return x

        self.mask = (np.random.rand(*x.shape) > self.p) / (1 - self.p)

        return x * self.mask

    def backward(self, grad):

        if not self.training:
            return grad

        return grad * self.mask



# 3. Softmax

class Softmax(Layer):

    def forward(self, x):

        x_shifted = x - np.max(x, axis=1, keepdims=True)
        exp_x = np.exp(x_shifted)

        self.out = exp_x / np.sum(exp_x, axis=1, keepdims=True)

        return self.out

    def backward(self, grad):

        batch_size, n_classes = grad.shape
        dx = np.zeros_like(grad)

        for i in range(batch_size):

            y = self.out[i].reshape(-1, 1)

            jacobian = np.diagflat(y) - y @ y.T

            dx[i] = jacobian @ grad[i]

        return dx



# 4. Batch Normalization

class BatchNorm(Layer):

    def __init__(self, num_features, eps=1e-5, momentum=0.9):

        self.gamma = np.ones((1, num_features))
        self.beta = np.zeros((1, num_features))

        self.dgamma = np.zeros_like(self.gamma)
        self.dbeta = np.zeros_like(self.beta)

        self.eps = eps
        self.momentum = momentum

        self.running_mean = np.zeros((1, num_features))
        self.running_var = np.ones((1, num_features))

        self.training = True

    def forward(self, x):

        if self.training:

            self.x = x

            self.mean = np.mean(x, axis=0, keepdims=True)
            self.var = np.var(x, axis=0, keepdims=True)

            self.x_hat = (x - self.mean) / np.sqrt(self.var + self.eps)

            out = self.gamma * self.x_hat + self.beta

            self.running_mean = self.momentum * self.running_mean + (1 - self.momentum) * self.mean
            self.running_var = self.momentum * self.running_var + (1 - self.momentum) * self.var

            return out

        else:

            x_hat = (x - self.running_mean) / np.sqrt(self.running_var + self.eps)

            return self.gamma * x_hat + self.beta

    def backward(self, grad):

        m = grad.shape[0]

        dbeta = np.sum(grad, axis=0, keepdims=True)
        dgamma = np.sum(grad * self.x_hat, axis=0, keepdims=True)

        dx_hat = grad * self.gamma

        dvar = np.sum(dx_hat * (self.x - self.mean) * -0.5 * (self.var + self.eps)**(-3/2), axis=0, keepdims=True)

        dmean = np.sum(dx_hat * -1 / np.sqrt(self.var + self.eps), axis=0, keepdims=True) + \
                dvar * np.mean(-2 * (self.x - self.mean), axis=0, keepdims=True)

        dx = dx_hat / np.sqrt(self.var + self.eps) + \
             dvar * 2 * (self.x - self.mean) / m + \
             dmean / m

        self.dgamma = dgamma
        self.dbeta = dbeta

        return dx


import numpy as np


class Sequential:
    """
    Container for stacking layers sequentially.
    """

    def __init__(self, *layers):
        self.layers = list(layers)
    
    def __call__(self, x):
        return self.forward(x)
    
    def __len__(self):
        return len(self.layers)


    # Forward Pass

    def forward(self, x):

        for layer in self.layers:
            x = layer.forward(x)

        return x


    # Backward Pass

    def backward(self, grad):

        for layer in reversed(self.layers):
            grad = layer.backward(grad)

        return grad

    
    # Collect Parameters

    def parameters(self):

        params = []

        for layer in self.layers:

            for attr in dir(layer):

                if attr.startswith("d"):  
                    param_name = attr[1:]   # remove "d"
                    grad_name = attr

                    if hasattr(layer, param_name):

                        params.append(
                            (layer, param_name, grad_name)
                        )

        return params

    
    # Zero Gradients

    def zero_grad(self):

        for layer, param_name, grad_name in self.parameters():

            grad = getattr(layer, grad_name)

            grad.fill(0)


    # Training Mode

    def train(self):

        for layer in self.layers:
            if hasattr(layer, "training"):
                layer.training = True

    
    # Evaluation Mode

    def eval(self):

        for layer in self.layers:
            if hasattr(layer, "training"):
                layer.training = False


    # Add Layer Dynamically

    def add(self, layer):

        self.layers.append(layer)

    
    # Model Summary

    def summary(self):

        print("Model Summary")
        print("-" * 40)

        for i, layer in enumerate(self.layers):
            print(f"{i+1}. {layer.__class__.__name__}")

        print("-" * 40)


import numpy as np


class Optimizer:
    """Base optimizer class"""

    def __init__(self, parameters, lr=0.01):
        self.parameters = parameters
        self.lr = lr

    def step(self):
        raise NotImplementedError

    def zero_grad(self):

        for layer, param_name, grad_name in self.parameters:
            grad = getattr(layer, grad_name)
            grad.fill(0)



# 1. Stochastic Gradient Descent

class SGD(Optimizer):

    def step(self):

        for layer, param_name, grad_name in self.parameters:

            param = getattr(layer, param_name)
            grad = getattr(layer, grad_name)

            param -= self.lr * grad



# 2. Momentum

class Momentum(Optimizer):

    def __init__(self, parameters, lr=0.01, beta=0.9):
        super().__init__(parameters, lr)
        self.beta = beta

        self.v = []
        for layer, param_name, _ in parameters:
            param = getattr(layer, param_name)
            self.v.append(np.zeros_like(param))

    def step(self):

        for i, (layer, param_name, grad_name) in enumerate(self.parameters):

            param = getattr(layer, param_name)
            grad = getattr(layer, grad_name)

            self.v[i] = self.beta * self.v[i] + (1 - self.beta) * grad

            param -= self.lr * self.v[i]



# 3. Nesterov Accelerated Gradient

class Nesterov(Optimizer):

    def __init__(self, parameters, lr=0.01, beta=0.9):
        super().__init__(parameters, lr)
        self.beta = beta

        self.v = []
        for layer, param_name, _ in parameters:
            param = getattr(layer, param_name)
            self.v.append(np.zeros_like(param))

    def step(self):

        for i, (layer, param_name, grad_name) in enumerate(self.parameters):

            param = getattr(layer, param_name)
            grad = getattr(layer, grad_name)

            v_prev = self.v[i]

            self.v[i] = self.beta * self.v[i] + self.lr * grad

            param -= (-self.beta * v_prev + (1 + self.beta) * self.v[i])



# 4. AdaGrad

class AdaGrad(Optimizer):

    def __init__(self, parameters, lr=0.01, eps=1e-8):
        super().__init__(parameters, lr)
        self.eps = eps

        self.G = []
        for layer, param_name, _ in parameters:
            param = getattr(layer, param_name)
            self.G.append(np.zeros_like(param))

    def step(self):

        for i, (layer, param_name, grad_name) in enumerate(self.parameters):

            param = getattr(layer, param_name)
            grad = getattr(layer, grad_name)

            self.G[i] += grad ** 2

            param -= self.lr * grad / (np.sqrt(self.G[i]) + self.eps)



# 5. RMSProp

class RMSProp(Optimizer):

    def __init__(self, parameters, lr=0.001, beta=0.9, eps=1e-8):
        super().__init__(parameters, lr)
        self.beta = beta
        self.eps = eps

        self.s = []
        for layer, param_name, _ in parameters:
            param = getattr(layer, param_name)
            self.s.append(np.zeros_like(param))

    def step(self):

        for i, (layer, param_name, grad_name) in enumerate(self.parameters):

            param = getattr(layer, param_name)
            grad = getattr(layer, grad_name)

            self.s[i] = self.beta * self.s[i] + (1 - self.beta) * grad ** 2

            param -= self.lr * grad / (np.sqrt(self.s[i]) + self.eps)



# 6. Adam

class Adam(Optimizer):

    def __init__(self, parameters, lr=0.001, beta1=0.9, beta2=0.999, eps=1e-8):

        super().__init__(parameters, lr)

        self.beta1 = beta1
        self.beta2 = beta2
        self.eps = eps

        self.m = []
        self.v = []

        for layer, param_name, _ in parameters:
            param = getattr(layer, param_name)
            self.m.append(np.zeros_like(param))
            self.v.append(np.zeros_like(param))

        self.t = 0

    def step(self):

        self.t += 1

        for i, (layer, param_name, grad_name) in enumerate(self.parameters):

            param = getattr(layer, param_name)
            grad = getattr(layer, grad_name)

            self.m[i] = self.beta1 * self.m[i] + (1 - self.beta1) * grad
            self.v[i] = self.beta2 * self.v[i] + (1 - self.beta2) * (grad ** 2)

            m_hat = self.m[i] / (1 - self.beta1 ** self.t)
            v_hat = self.v[i] / (1 - self.beta2 ** self.t)

            param -= self.lr * m_hat / (np.sqrt(v_hat) + self.eps)


import numpy as np


class Loss:
    """Base class for all loss functions."""

    def forward(self, y_pred, y_true):
        raise NotImplementedError

    def backward(self):
        raise NotImplementedError



# 1. Mean Squared Error

class MeanSquaredError(Loss):

    def forward(self, y_pred, y_true):
        self.y_pred = y_pred
        self.y_true = y_true
        return np.mean((y_pred - y_true) ** 2)

    def backward(self):
        n = self.y_true.size
        return 2 * (self.y_pred - self.y_true) / n



# 2. Mean Absolute Error

class MeanAbsoluteError(Loss):

    def forward(self, y_pred, y_true):
        self.y_pred = y_pred
        self.y_true = y_true
        return np.mean(np.abs(y_pred - y_true))

    def backward(self):
        n = self.y_true.size
        return np.sign(self.y_pred - self.y_true) / n



# 3. Log Loss (Binary Cross Entropy)

class LogLoss(Loss):

    def forward(self, y_pred, y_true):

        eps = 1e-15
        y_pred = np.clip(y_pred, eps, 1 - eps)

        self.y_pred = y_pred
        self.y_true = y_true

        loss = -np.mean(
            y_true * np.log(y_pred) +
            (1 - y_true) * np.log(1 - y_pred)
        )

        return loss

    def backward(self):

        n = self.y_true.shape[0]

        grad = (
            (self.y_pred - self.y_true) /
            (self.y_pred * (1 - self.y_pred))
        ) / n

        return grad



# 4. Cross Entropy Loss (Multi-class)

class CrossEntropyLoss(Loss):

    def forward(self, y_pred, y_true):

        eps = 1e-15
        y_pred = np.clip(y_pred, eps, 1 - eps)

        self.y_pred = y_pred
        self.y_true = y_true

        n = y_true.shape[0]

        loss = -np.sum(y_true * np.log(y_pred)) / n

        return loss

    def backward(self):

        n = self.y_true.shape[0]

        grad = - (self.y_true / self.y_pred) / n

        return grad



# 5. Huber Loss

class HuberLoss(Loss):

    def __init__(self, delta=1.0):
        self.delta = delta

    def forward(self, y_pred, y_true):

        self.y_pred = y_pred
        self.y_true = y_true

        error = y_pred - y_true
        abs_error = np.abs(error)

        quadratic = np.minimum(abs_error, self.delta)
        linear = abs_error - quadratic

        loss = np.mean(
            0.5 * quadratic ** 2 +
            self.delta * linear
        )

        return loss

    def backward(self):

        error = self.y_pred - self.y_true

        grad = np.where(
            np.abs(error) <= self.delta,
            error,
            self.delta * np.sign(error)
        )

        return grad / self.y_true.size



# 6. Hinge Loss (SVM)

class HingeLoss(Loss):

    def forward(self, y_pred, y_true):

        # y_true must be -1 or +1

        self.y_pred = y_pred
        self.y_true = y_true

        loss = np.mean(
            np.maximum(0, 1 - y_true * y_pred)
        )

        return loss

    def backward(self):

        grad = np.zeros_like(self.y_pred)

        mask = (1 - self.y_true * self.y_pred) > 0

        grad[mask] = -self.y_true[mask]

        return grad / self.y_true.shape[0]